import pytest
from gcsfs.tests.settings import TEST_BUCKET
from gcsfs.tests.utils import my_vcr, gcs_maker

root = TEST_BUCKET + "/mapping"


pytestmark = pytest.mark.usefixtures("token_restore")


def test_api():
    import gcsfs

    assert "GCSMap" in dir(gcsfs)
    assert "mapping" in dir(gcsfs)


@my_vcr.use_cassette(match=["all"])
def test_map_simple():
    with gcs_maker() as gcs:
        d = gcs.get_mapper(root)
        assert not d

        assert list(d) == list(d.keys()) == []
        assert list(d.values()) == []
        assert list(d.items()) == []


@my_vcr.use_cassette(match=["all"])
def test_map_default_gcsfilesystem():
    with gcs_maker() as gcs:
        d = gcs.get_mapper(root)
        assert d.fs is gcs


@my_vcr.use_cassette(match=["all"])
def test_map_errors():
    with gcs_maker() as gcs:
        d = gcs.get_mapper(root)
        with pytest.raises(KeyError):
            d["nonexistent"]
        try:
            gcs.get_mapper("does-not-exist")
        except Exception as e:
            assert "does-not-exist" in str(e)


@pytest.mark.xfail(reason="only passes for the py version where it was recorded")
@my_vcr.use_cassette(match=["all"])
def test_map_with_data():
    with gcs_maker() as gcs:
        d = gcs.get_mapper(root)
        d["x"] = b"123"
        assert list(d) == list(d.keys()) == ["x"]
        assert list(d.values()) == [b"123"]
        assert list(d.items()) == [("x", b"123")]
        assert d["x"] == b"123"
        assert bool(d)

        assert gcs.find(root) == [TEST_BUCKET + "/mapping/x"]
        d["x"] = b"000"
        assert d["x"] == b"000"

        d["y"] = b"456"
        assert d["y"] == b"456"
        assert set(d) == {"x", "y"}

        d.clear()
        assert list(d) == []


@my_vcr.use_cassette(match=["all"])
def test_map_complex_keys():
    with gcs_maker() as gcs:
        d = gcs.get_mapper(root)
        d[1] = b"hello"
        assert d[1] == b"hello"
        del d[1]

        d[1, 2] = b"world"
        assert d[1, 2] == b"world"
        del d[1, 2]

        d["x", 1, 2] = b"hello world"
        assert d["x", 1, 2] == b"hello world"

        assert ("x", 1, 2) in d


@my_vcr.use_cassette(match=["all"])
def test_map_clear_empty():
    with gcs_maker() as gcs:
        d = gcs.get_mapper(root)
        d.clear()
        assert list(d) == []
        d[1] = b"1"
        # may repeat the test below, since VCR sometimes picks the wrong call to ls
        assert list(d) == ["1"] or list(d) == ["1"]
        d.clear()
        assert list(d) == []


@my_vcr.use_cassette(match=["all"])
def test_map_pickle():
    with gcs_maker() as gcs:
        d = gcs.get_mapper(root)
        d["x"] = b"1"
        assert d["x"] == b"1"

        import pickle

        d2 = pickle.loads(pickle.dumps(d))

        assert d2["x"] == b"1"


@my_vcr.use_cassette(match=["all"])
def test_map_array():
    with gcs_maker() as gcs:
        from array import array

        d = gcs.get_mapper(root)
        d["x"] = array("B", [65] * 1000)

        assert d["x"] == b"A" * 1000


@my_vcr.use_cassette(match=["all"])
def test_map_bytearray():
    with gcs_maker() as gcs:
        d = gcs.get_mapper(root)
        d["x"] = bytearray(b"123")

        assert d["x"] == b"123"


@my_vcr.use_cassette(match=["all"])
def test_new_bucket():
    with gcs_maker() as gcs:
        new_bucket = TEST_BUCKET + "new-bucket"
        try:
            gcs.rmdir(new_bucket)
        except:  # noqa: E722
            pass
        with pytest.raises(Exception) as e:
            d = gcs.get_mapper(new_bucket, check=True)
        assert "create=True" in str(e.value)

        try:
            d = gcs.get_mapper(new_bucket, create=True)
            assert not d

            d = gcs.get_mapper(new_bucket + "/new-directory")
            assert not d
        finally:
            gcs.rmdir(new_bucket)


@my_vcr.use_cassette(match=["all"])
def test_map_pickle():
    import pickle

    with gcs_maker() as gcs:
        d = gcs.get_mapper(root)
        d["x"] = b"1234567890"

        b = pickle.dumps(d)
        assert b"1234567890" not in b

        e = pickle.loads(b)

        assert dict(e) == {"x": b"1234567890"}



import pytest

from gcsfs.core import GCSFileSystem, GoogleCredentials
from gcsfs.tests.settings import TEST_PROJECT, TEST_BUCKET
import vcr.stubs.aiohttp_stubs as aios


import fsspec.config

fsspec.config.conf.pop("gcs", None)


@pytest.fixture
def token_restore():
    cache = GoogleCredentials.tokens
    try:
        GoogleCredentials.tokens = {}
        yield
    finally:
        GoogleCredentials.tokens = cache
        GoogleCredentials._save_tokens()
        GCSFileSystem.clear_instance_cache()


# patch; for some reason, original wants vcr_response["url"], which is empty
def build_response(vcr_request, vcr_response, history):
    request_info = aios.RequestInfo(
        url=aios.URL(vcr_request.url),
        method=vcr_request.method,
        headers=aios.CIMultiDictProxy(aios.CIMultiDict(vcr_request.headers)),
        real_url=aios.URL(vcr_request.url),
    )
    response = aios.MockClientResponse(
        vcr_request.method, aios.URL(vcr_request.url), request_info=request_info
    )
    response.status = vcr_response["status"]["code"]
    response._body = vcr_response["body"].get("string", b"")
    response.reason = vcr_response["status"]["message"]
    head = {
        k: v[0] if isinstance(v, list) else v
        for k, v in vcr_response["headers"].items()
    }
    response._headers = aios.CIMultiDictProxy(aios.CIMultiDict(head))
    response._history = tuple(history)

    response.close()
    return response


aios.build_response = build_response


# patch: but the value of body back in the stream, to enable streaming reads
# https://github.com/kevin1024/vcrpy/pull/509
async def record_response(cassette, vcr_request, response):
    """Record a VCR request-response chain to the cassette."""

    try:
        byts = await response.read()
        body = {"string": byts}
        if byts:
            if response.content._buffer_offset:
                response.content._buffer[0] = response.content._buffer[0][
                    response.content._buffer_offset :
                ]
                response.content._buffer_offset = 0
            response.content._size += len(byts)
            response.content._cursor -= len(byts)
            response.content._buffer.appendleft(byts)
            response.content._eof_counter = 0

    except aios.ClientConnectionError:
        body = {}

    vcr_response = {
        "status": {"code": response.status, "message": response.reason},
        "headers": aios._serialize_headers(response.headers),
        "body": body,  # NOQA: E999
        "url": str(response.url)
        .replace(TEST_BUCKET, "gcsfs-testing")
        .replace(TEST_PROJECT, "test_project"),
    }

    cassette.append(vcr_request, vcr_response)


aios.record_response = record_response


def play_responses(cassette, vcr_request):
    history = []
    vcr_response = cassette.play_response(vcr_request)
    response = build_response(vcr_request, vcr_response, history)

    # If we're following redirects, continue playing until we reach
    # our final destination.
    while 300 <= response.status <= 399:
        if "Location" not in response.headers:
            # Not a redirect, an instruction not to call again
            break
        next_url = aios.URL(response.url).with_path(response.headers["location"])

        # Make a stub VCR request that we can then use to look up the recorded
        # VCR request saved to the cassette. This feels a little hacky and
        # may have edge cases based on the headers we're providing (e.g. if
        # there's a matcher that is used to filter by headers).
        vcr_request = aios.Request(
            "GET",
            str(next_url),
            None,
            aios._serialize_headers(response.request_info.headers),
        )
        vcr_request = cassette.find_requests_with_most_matches(vcr_request)[0][0]

        # Tack on the response we saw from the redirect into the history
        # list that is added on to the final response.
        history.append(response)
        vcr_response = aios.cassette.play_response(vcr_request)
        response = aios.build_response(vcr_request, vcr_response, history)

    return response


aios.play_responses = play_responses



import json
import os
import gcsfs.core
import gcsfs.credentials

RECORD_MODE = os.environ.get("GCSFS_RECORD_MODE", "none")
TEST_PROJECT = os.environ.get("GCSFS_TEST_PROJECT", "test_project")

TEST_BUCKET = os.environ.get("GCSFS_TEST_BUCKET", "gcsfs-testing")
TEST_REQUESTER_PAYS_BUCKET = os.environ.get(
    "GCSFS_TEST_REQUESTER_PAYS_BUCKET", "gcsfs-testing-requesterpays"
)

FAKE_TOKEN = {
    "access_token": "xxx",
    "expires_in": 0,
    "grant_type": "refresh_token",
    "refresh_token": "xxx",
    "timestamp": 1487859400.0,
}

FAKE_TOKEN.update(gcsfs.credentials.not_secret)

FAKE_GOOGLE_TOKEN = {
    "client_id": (
        "764086051850-6qr4p6gpi6hn506pt8ejuq83di341hur." "apps.googleusercontent.com"
    ),
    "client_secret": "d-FL95Q19q7MQmFpd7hHD0Ty",
    "refresh_token": "xxx",
    "type": "authorized_user",
}
GOOGLE_TOKEN = os.environ.get("GCSFS_GOOGLE_TOKEN", FAKE_GOOGLE_TOKEN)
ON_VCR = "GCSFS_GOOGLE_TOKEN" not in os.environ

if isinstance(GOOGLE_TOKEN, str) and os.path.exists(GOOGLE_TOKEN):
    with open(GOOGLE_TOKEN) as f:
        GOOGLE_TOKEN = json.load(f)



from gcsfs.credentials import GoogleCredentials
from gcsfs.tests.utils import my_vcr


@my_vcr.use_cassette(match=["all"])
def test_GoogleCredentials_None():
    credentials = GoogleCredentials(project=None, token=None, access="read_only")
    headers = {}
    credentials.apply(headers)
